/* rstd v1.1, (c) 2000 Robert Salizar (madcamel@energymech.net)
 * This code is distributed under the propritary MCL license,
 * please read LICENSE
 */
#include "rstd.h"

void filter (char *packet);
void send_rst(ulong seq, ulong dst, int sport, int dport);
void daemonise();
int  loadcfg(char *cfgfile);
void handle_sighup();

int	sock;
int	rst_all	= 0;
uint	ports[65535];
int	cfglines;
time_t	last;

int main(int argc, char **argv) {
	char packet[44];

	if (argv[1]) {
		if (!strncmp(argv[1],  "-a", 2))
			rst_all = 1;
		else {
			fprintf(stderr, "usage: %s [-a]\n", argv[0]);
			exit(255);
		}
	}

	signal(SIGHUP, handle_sighup);

	if (!rst_all) {
		cfglines = loadcfg(CFGFILE);
	}		


	if ((sock = socket(PF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
		PERR("socket");

	if ((setsockopt(sock, IPPROTO_IP, IP_HDRINCL, "1", 1)) < 0)
		PERR("setsockopt");

	daemonise();

/* GOTO??! you ask? Why not? It shaves a few instructions out of the 
 * loop and makes the code look cleaner. */
loop:
		read(sock, packet, 44);
		filter(packet);
		goto loop;
}

void filter (char *packet)
{
	/* Easier to use iphdr instead of ip here */
	struct iphdr	*ih = (struct iphdr *) packet;
	struct tcphdr	*th = (struct tcphdr *) (packet + sizeof(struct iphdr));
	int    x;

	if (th->th_flags != (TH_SYN))
		return;
	
	/* Throttle to 1 rst/sec */
	if (last == time(NULL))
		return;

	last = time(NULL);

	if (rst_all) {
		send_rst(ntohl(th->th_seq), ih->saddr, th->th_dport, 
			th->th_sport);
		return;
	}

	for(x = 0; x < cfglines; x++) {
		if (th->th_dport == htons(ports[x])) {
			send_rst(ntohl(th->th_seq), ih->saddr, 
				th->th_dport, th->th_sport);
			break;
		}
	}
}

/* RST packets still need checksum. Psuedoheader is yucky. */
void send_rst(ulong seq, ulong dst, int sport, int dport)
{
	struct packet		packet;
	struct cksum		cksum;
	struct sockaddr_in	s_in;

	memset(&packet, 0, sizeof(packet));

	/* IP */
	packet.ip.ip_hl		= 5;
	packet.ip.ip_v		= 4;
	packet.ip.ip_p		= IPPROTO_TCP;
	packet.ip.ip_tos	= 0;
	packet.ip.ip_id		= rand();
	packet.ip.ip_len	= htons(sizeof(packet));
	packet.ip.ip_off	= 0;
	packet.ip.ip_ttl	= 255;
	packet.ip.ip_dst.s_addr	= dst;

	/* TCP */		/* Yes this is right */
	packet.tcp.th_flags	= (TH_RST | TH_ACK);
	packet.tcp.th_win 	= 0;
	packet.tcp.th_seq	= 0;
	packet.tcp.th_off	= 21;
	packet.tcp.th_urp	= 0;
	packet.tcp.th_ack	= htonl(++seq);
	packet.tcp.th_sport	= sport;
	packet.tcp.th_dport	= dport;
	packet.opt.len		= 0;

	/* Happy fun psuedoheader */
	cksum.pseudo.daddr	= dst;
	cksum.pseudo.mbz	= 0;
	cksum.pseudo.ptcl	= IPPROTO_TCP;
	cksum.pseudo.tcpl	= htons(sizeof(struct tcphdr));
	
	/* Good ol sockaddr_in */
	s_in.sin_family		= AF_INET;
	s_in.sin_addr.s_addr	= dst;
	s_in.sin_port		= packet.tcp.th_dport;

	/* Checksum, yum. */
	cksum.tcp		= packet.tcp;
	packet.ip.ip_sum	= in_cksum ((void *)&packet.ip, 20);
	packet.tcp.th_sum	= in_cksum ((void *)&cksum, sizeof(cksum));

	/* Blast off! */
	sendto (sock, &packet, sizeof(packet) - 4, 0, 
		(struct sockaddr *)&s_in, sizeof(s_in));
}

void daemonise() {
	int forked;

	if ((forked = fork()) < 0)
		PERR("fork");

	if (forked)
		_exit(0);

	if (setsid() < 0)
		PERR("setsid");

	chdir("/");
	close(0);
	close(1);
	close(2);
}

int loadcfg(char *cfgfile) {
	FILE	*cfg;
	char	buf[8];
	int	x;

	cfg = fopen(cfgfile, "r");
	if (!cfg)
		PERR(cfgfile);

	for (x = 0; ((fgets(buf, 7, cfg) != NULL) && (x < 65535)); x++) 
{
		if (buf[0] == '#' || buf[0] == '\n') {
			x--;
			continue;
		}
		buf[strlen(buf) - 1] = 0;

		if (!atoi(buf)) {
			fprintf(stderr, "%s: bad line: %d\n", cfgfile, ++x);
			exit(1);
		}

		ports[x] = atoi(buf);
	}
	return(x);
}

void handle_sighup() {
	if (!rst_all)
		cfglines = loadcfg(CFGFILE);
}
